/*--------------------------------------------------------------------------
     "combust.cc" - generate combustion mechanism
        This is a part of KUCRS:
           Knowledge-basing Utilities for Complex Reaction Systems
                        (c) 2002-2021 by A. Miyoshi, Hiroshima University
                                       created: Aug.   5, 2002
                                        edited: Aug.  21, 2004
                                        edited: Jan.  26, 2011
                                   last edited: Aug.  22, 2021
--------------------------------------------------------------------------*/

#include "reaction.h"
using namespace std;

#define FL_MOLMAI "mol_main.dat"        // main molecule library

#define FN_OUTTHG "_thg.inp"            // thermgen input
#define FN_OUTCHG "_chg.inp"            // chemgen input
#define FN_OUTMOL "_mol.dat"            // molecle library output
#define FN_OUTMLL "_mol.log"            // molecle library log
#define FN_OUTSNK "_snk.inp"            // sample senkin input
#define FN_OUTTHF "_thf.inp"            // thermfit input
#define FN_OUTFUE "_fue.log"            // fuel output
#define FN_COMTHG "_thg"                // common thermgen name

#define FI_TRAD   0                     // fuel input - traditional
#define FI_SMILES 1                     // fuel input - SMILES

int readFuel(string flname, vector<molecule> &fuels, libPath &libp,
 groupLib &gl, typeMolec &tm);
int sepMultiFuels(string s, vector<string> &finps, vector<int> &finpTyp);
void procFuels(vector<molecule> &fuels, specAdmin &spad, outStrms &ofs,
 groupLib &grp, typeMolec &tym);
int outSampSenkInp(string flname, vector<molecule> &fuels);
int outThmfitInp(string basNam, string addNfit, string addNgen);

/*===== main function ====================================================*/

int main(int argc, char *argv[]) {
  int rxnError = false;
  string cdate, fnInp, fnBas, flnFuelOut;
  ofstream osfuel;
  libPath libp;
  groupLib grp;
  specAdmin spad;
  rxnLib rxl;
  rxnTunePar rtu;
  typeMolec tym;
  outStrms ofs;
  vector<molecule> fmols;
  vector<molecule>::iterator pfu;
  reaction rxn;

  if (argc != 2) { cout << "usage: combust input\n"; return 1; }
  cdate = "(" + strCurDate() + ")";
  cout << "--- Program combust " << cdate << " ---\n";
  spad.clear();
  if (spad.molAd.appendFrom(FL_MOLMAI, libp, spad.namAd, grp) == ERRE) {
    cout << "Failed to read molecular library.\n";
    return 1;
  }
  fnInp = argv[1]; fnBas = getBaseName(fnInp);
  if (readFuel(fnInp, fmols, libp, grp, tym) == ERRE)
    { cout << "Error reading fuel input.\n"; return 1; }
  if (ofs.open(fnBas + FN_OUTTHG, fnBas + FN_OUTCHG, fnBas + FN_OUTMOL, fnBas + FN_OUTMLL)
      == ERRE) { return 1; }
  flnFuelOut = fnBas + FN_OUTFUE;
  osfuel.open(flnFuelOut.c_str());
  if (!osfuel) { cout << "Error writing " << flnFuelOut << ".\n"; return 1; }
  ofs.thm << "Thermodata generated by combust " << cdate << endl;
  ofs.rxn << "! Reaction data generated by combust " << cdate << endl;
  procFuels(fmols, spad, ofs, grp, tym);
  for (pfu = fmols.begin(); pfu != fmols.end(); pfu++)
   { osfuel << pfu->molName << ' ' << pfu->getSMILES() << endl; }
  rxn.initPrivMemb(spad, grp, rxl, rtu, tym, ofs, fmols);
  while (spad.rxnQue.size() != 0) {
    if (rxn.reactQTop() == ERRE)
      { cout << "Error generating reactions.\n"; rxnError = true; break; }
  }
  ofs.close();
  osfuel.close();
  if (rxnError) { return 1; }
  if (outSampSenkInp(fnBas + FN_OUTSNK, fmols) == ERRE) { return 1; }
  if (outThmfitInp(fnBas, FN_OUTTHF, FN_COMTHG) == ERRE) { return 1; }
  rxn.reportGenStat();
  return 0;
}

/*----- read fuels from an input file ------------------------------------*/

int readFuel(string flname, vector<molecule> &fuels, libPath &libp,
 groupLib &gl, typeMolec &tm) {
  int ix, iy, isDupl, isSupp;
  long mtyp;
  string s, typFuel;
  vector<int> finpTyp;       // fuel input type; FI_TRAD or FI_SMILES
  vector<string> finps, vs;
  vector<vector<int> > dblBnds;
  alkaneChain alc;
  molecule fmol;

  if (getCharsFile(flname, s) == ERRE) {
    cout << "failed to read HC from [" << flname << "].\n";
    return ERRE;
  }
  if (sepMultiFuels(s, finps, finpTyp) == ERRE)
    { cout << "Wrong fuel input.\n"; return ERRE; }
  if (!fuels.empty()) { fuels.clear(); }
  for (ix = 0; ix < finps.size(); ix++) {
    if (finpTyp[ix] == FI_TRAD) {
      alc.clear(); if (alc.setFrom(finps[ix]) == ERRE) { return ERRE; }
      fmol.clear(); if (fmol.setFrom(alc) == ERRE) { return ERRE; }
    } else {
      fmol.clear();
      if (fmol.setFromPsmiles(finps[ix], gl, tm) == ERRE) { return ERRE; }
    }
    cout << "formula : " << fmol.molec.getChemFormula() << endl;
    cout << "name : " << fmol.molName << endl;
    fmol.beUnique(); fmol.getUniqueView(vs);
    cout << "structure :\n" << vs;
    mtyp = fmol.getType(gl);
    typFuel = tm.typName(mtyp);
    cout << "type: " << typFuel << " (" << hex << mtyp << dec << ")\n";
    isSupp = false;
    if ((typFuel == "Alkane") || (typFuel == "cyAlkane")) { isSupp = true; }
    if (typFuel == "Alkene") {
      dblBnds = fmol.molec.getAllBondsMatch(6, 6, 2);
      if (dblBnds.size() == 1) { isSupp = true; }
    }
    if (typFuel == "Alcohol") { isSupp = true; }
    if (!isSupp) {
      cout << "unsupprted fuel type\n";
      cout << "supported fules are:\n"
           << " - Alkanes (non- and mono-cyclic)\n"
           << " - Alkenes (non-cyclic mono-ens)\n"
           << " - Alkohols (non-cylic)\n";
      return ERRE;
    }
    isDupl = false;
    for (iy = 0; iy < fuels.size(); iy++) {
      if (fmol.molec == fuels[iy].molec) {
        isDupl = true;
        cout << "! Ignored duplicated fuel input.\n";
        break;
      }
    }
    if (!isDupl) { fuels.push_back(fmol); }
  }
  return NORM;
}

/*----- separate multiple-fuel input --------------------------------*/

int sepMultiFuels(string s, vector<string> &finps, vector<int> &finpTyp) {
  int slen, p = 0, pst, ped, pad;
  string salk, ssml;

  if (!finps.empty()) { finps.clear(); }
  if (!finpTyp.empty()) { finpTyp.clear(); }
  slen = s.length();
  while (p < (slen - 6)) {
    if (s.substr(p, 6) == "alkane") {
      if (s[p + 6] == '{') { pst = p + 7; ped = s.find('}', pst); pad = 1; }
      else { pst = p + 6; ped = s.find("end_alkane", pst); pad = 10; }
      if (ped == string::npos) { return ERRE; }
      salk = s.substr(pst, ped - pst);
      if (salk != "") { finps.push_back(salk); finpTyp.push_back(FI_TRAD); }
      p = ped + pad;
    } else if (s.substr(p, 5) == "fuel{") {
      pst = p + 5;
      if ((ped = s.find('}', pst)) == string::npos) { return ERRE; }
      ssml = s.substr(pst, ped - pst);
      if (ssml != "") { finps.push_back(ssml); finpTyp.push_back(FI_SMILES); }
      p = ped + 1;
    }
  }
  if (finps.size() == 0) { return ERRE; }
  return NORM;
}

/*----- process fuels ----------------------------------------------------*/

void procFuels(vector<molecule> &fuels, specAdmin &spad, outStrms &ofs,
 groupLib &grp, typeMolec &tym) {
  int ix, mstat;
  string namf;

  for (ix = 0; ix < fuels.size(); ix++) {
    mstat = spad.getMolStat(fuels[ix], namf);
    if (mstat != 3) {
      if (mstat == 2) {
        fuels[ix].molName = namf; spad.molAd.removeMol(fuels[ix]);
        cout << "fuel [" << namf << "] found in lib but rxn not found.";
      } else {
        namf = spad.namAd.regBaseSfx(fuels[ix].molName, "");
        fuels[ix].molName = namf;
        cout << "fuel [" << namf << "] not found in lib.";
      }
      ofs.thm << fuels[ix].getThermInput(grp) << endl;
      fuels[ix].storeTo(ofs.mol);
      fuels[ix].storeLogTo(ofs.mll, grp, tym);
      spad.rxnQue.push_back(molecInQue(fuels[ix]));
      fuels[ix].isRxnDone = false;
      cout << " - full rxn will be generated.\n";
    } else {
      fuels[ix].molName = namf;
      fuels[ix].isRxnDone = true;
      cout << "fuel [" << namf << "] & its rxn found in lib.";
      cout << " - reaction of new-RO2 + fuel";
      if (namf != "CH4") {
        cout << ", new-RO2 + fuel-R, and new-RO2 + fuel-RO2";
      }
      cout << " will be generated.\n";
    }
  }
}

/*----- output sample senkin input ---------------------------------------*/

int outSampSenkInp(string flname, vector<molecule> &fuels) {
  int ix;
  double amntO2, amntN2;
  ofstream ofs;
  atmcComp ac;
  atmcComp::iterator p;

  ofs.open(flname.c_str());
  if (!ofs) { cout << "Failed to write " << flname << ".\n"; return ERRE; }
  printHeadingInfo(ofs);
  ofs << "CONV\nPRES 40.0\nTEMP 750.\nTIME 10.E-3\nDELT 5.E-4\n";
  for (ix = 0; ix < fuels.size(); ix++) {
    fuels[ix].molec.getAtomComp(ac); amntO2 = 0.;
    for (p = ac.begin(); p != ac.end(); p++) {
      if (p->first == 1) { amntO2 += 0.25 * p->second;  }
      else if (p->first == 6) { amntO2 += 1.0 * p->second; }
      else if (p->first == 8) { amntO2 -= 0.5 * p->second; }
    }
    amntN2 = amntO2 * 79.05 / 20.95;
    if (ix == 0) {
      ofs << "REAC " << fuels[ix].molName << " 1.\n";
      ofs << "REAC O2 " << formatFautoLen(amntO2, 1) << endl;
      ofs << "REAC N2 " << formatFautoLen(amntN2, 6) << endl;
    } else {
      ofs << "! REAC " << fuels[ix].molName << " 1.\n";
      ofs << "! REAC O2 " << formatFautoLen(amntO2, 1) << endl;
      ofs << "! REAC N2 " << formatFautoLen(amntN2, 6) << endl;
    }
  }
  ofs << "END\n";
  ofs.close();
  cout << "senkin input sample generated in [" << flname << "].\n";
  return NORM;
}

/*----- output thermfit input --------------------------------------------*/

int outThmfitInp(string basNam, string addNfit, string addNgen) {
  string flname, thgname;
  ofstream ofs;

  flname = basNam + addNfit;
//  thgname = basNam + addNgen;
  thgname = addNgen;
  ofs.open(flname.c_str());
  if (!ofs) { cout << "Failed to write " << flname << ".\n"; return ERRE; }
    ofs << endl << endl << thgname << endl << thgname << endl;
  ofs.close();
  cout << "thermfit input generated in [" << flname << "].\n";
  return NORM;
}

