/*--------------------------------------------------------------------------
     "libfunc.cc" - library functions and classes
        This is a part of KUCRS:
           Knowledge-basing Utilities for Complex Reaction Systems
                        (c) 2002-2021 by A. Miyoshi, Hiroshima University
                                       created: Aug.   5, 2002
                                        edited: May    1, 2004
                                        edited: Apr.  14, 2005
                                        edited: Dec.  15, 2016
                                   last edited: Aug.  18, 2021
--------------------------------------------------------------------------*/

#include "libfunc.h"

// =========================================================================
//        general purpose functions and classes
// =========================================================================

/*===== string formatting and parsing functions ========================*/

string strToUpper(const string &s) {
  int ic;
  string rs = s;
  for(ic = 0; ic < rs.length(); ic++)
    { if (islower(rs[ic])) { rs[ic] = toupper(rs[ic]); } }
  return rs;
}

string strToLower(const string &s) {
  int ic;
  string rs = s;
  for(ic = 0; ic < rs.length(); ic++)
    { if (isupper(rs[ic])) { rs[ic] = tolower(rs[ic]); } }
  return rs;
}

string format(int v) {
  ostringstream os;
  os << v;
  return os.str();
}

string format(int v, int len) {
  ostringstream os;
  os << setw(len) << setfill('0') << v;
  return os.str();
}

string format(string s, int len) {
  ostringstream os;
  os << setw(len) << setfill(' ') << s.c_str();
  return os.str();
}

string formatLeft(string s, int len) {
  ostringstream os;
  os << setw(len) << setiosflags(ios::left) << s.c_str();
  return os.str();
}

string formatRight(string s, int len) {
  ostringstream os;
  os << setw(len) << setiosflags(ios::right) << s.c_str();
  return os.str();
}

string formatE(double v, int len, int prc) {
  ostringstream os;
  string rs;
  int posE;
  os << setiosflags(ios::right) << setiosflags(ios::showpoint)
     << setiosflags(ios::scientific) << setw(len) << setprecision(prc)
     << v;
  //--- for compiler that produces e+000 output ---
  rs = os.str(); posE = rs.find('e');
  if ((rs.length() - posE) > 4) {
    if (rs[posE + 2] == '0') {
      rs = rs.substr(0, posE + 2) + rs.substr(posE + 3);
      if (rs.length() < len) { rs = ' ' + rs; }
    }
  }
  return rs;
}

string formatF(double v, int len, int prc) {
  ostringstream os;
  os << setiosflags(ios::right) << setiosflags(ios::showpoint)
     << setiosflags(ios::fixed) << setw(len) << setprecision(prc)
     << v;
  return os.str();
}

string formatFnoDot(double v, int len) {
  ostringstream os;
  os << setiosflags(ios::right) << setiosflags(ios::fixed)
     << setw(len) << setprecision(0) << v;
  return os.str();
}

string formatFautoLen(double v, int prc) {
  ostringstream os;
  os << setiosflags(ios::right) << setiosflags(ios::showpoint)
     << setiosflags(ios::fixed) << setprecision(prc) << v;
  return os.str();
}

string formatD(int v, int len) {
  ostringstream os;
  os << setw(len) << setfill(' ') << v;
  return os.str();
}

string formatBool(int flag) {
  if (flag) { return "Y"; } else { return "N"; }
}

int valBoolStr(string s) {
  if ( (s == "Y") || (s == "y") || (s == "YES") || (s == "Yes") ||
       (s == "yes") ) { return true; }
  else { return false; }
}

long valHexStr(string s) {
  static const string hxtbl = "0123456789abcdef";
  char c; int ic, dig;
  long v = 0L;
  for (ic = 0; ic < s.length(); ic++) {
    c = s[ic]; if (isupper(c)) { c = tolower(c); }
    if ((dig = hxtbl.find(c)) == string::npos) { dig = 0; }
    v <<= 4; v += dig;
  }
  return v;
}

string strCurDate() {
char sdate[20];
time_t curTime;
  time(&curTime);
  strftime(sdate, (size_t)19, "%b %d, %Y", localtime(&curTime));
  return sdate;
}

int nextCorCket(string &s, int start) {
  int lev = 0, retPos = -1;
  for (int i = start; i < s.length(); i++) {
    if (s[i] == ')') {
      if (lev == 0) { retPos = i; break; }
      else { lev--; }
    } else if (s[i] == '(') {
      lev++;
    }
  }
  return retPos;
}

int numMatchStrList(string s, vector<string> *vs) {
  int nmat = 0;
  for (int i = 0; i < vs->size(); i++) {
    if ((*vs)[i] == s) { nmat++; }
  }
  return nmat;
}

ostream &operator<<(ostream &os, vector<string> &vs) {
  vector<string>::iterator p;
  for (p = vs.begin(); p != vs.end(); p++) { os << *p << endl; }
  return os;
}

//----- split tokens in string s ---------------------------------------

int splitToken(const string &s, const string &spl, vector<string> &vs) {
  char c;
  int pTkn, pLastTkn = -1, lens = s.length(), ic, lss;
  if (!vs.empty()) { vs.clear(); }
  for (ic = 0; ic < lens; ic++) {
    c = s[ic];
    if (spl.find(c) != string::npos) {
      pTkn = ic; lss = pTkn - pLastTkn - 1;
      if (lss > 0) { vs.push_back(s.substr(pLastTkn + 1, lss)); }
      pLastTkn = pTkn;
    }
  }
  if (pLastTkn < (lens - 1)) {
    pTkn = lens; lss = pTkn - pLastTkn - 1;
    vs.push_back(s.substr(pLastTkn + 1, lss));
  }
  return vs.size();
}

//----- string terminated by the specified char ------------------------

string strTermBy(const string &s, char ct) {
  int tpos;
  tpos = s.find(ct);
  return s.substr(0, tpos);
}

//----- remove preceding and subsequent space chars --------------------

string strRemoveSpc(const string &s) {
  int ic, pstart, pend, slen;
  string rs;
  slen = s.length(); pstart = slen; pend = -1;
  for (ic = 0; ic < slen; ic++) {
    if (s[ic] != ' ') { pstart = ic; break; }
  }
  for (ic = slen - 1; ic >= 0; ic--) {
    if (s[ic] != ' ') { pend = ic; break; }
  }
  if (pend < pstart) { rs = ""; }
  else { rs = s.substr(pstart, pend - pstart + 1); }
  return rs;
}

//----- get one line from the file stream ------------------------------

int getOneLine(ifstream &inf, string &s) {
  char c;
  if (s != "") { s = ""; }
  while (!inf.get(c).eof()) {
    if (c == '\n') { break; }
    if (!iscntrl(c)) { s += c; }
  }
  if (inf.eof() && (s == "")) { return ERRE; }
  return NORM;
}

//----- get net ntkn tokens from the file stream -----------------------

int getTokensFrom(ifstream &inf, int ntkn, vector<string> &vs) {
  int itk;
  string s;
  if (!vs.empty()) { vs.clear(); }
  for (itk = 0; itk < ntkn; itk++) {
    if (getTokenFrom(inf, s) == ERRE) { return ERRE; }
    vs.push_back(s);
  }
  return NORM;
}

/*----- get the next token from the file stream ------------------------
     ignoring comments ('!' to line-end)
----------------------------------------------------------------------*/

int getTokenFrom(ifstream &inf, string &s) {
  char c;
  int inComment = false, inToken = false, doneToken = false;
  static string tokenSep = " ,\t\n\r";
  s = "";
  if (inf.eof()) { return ERRE; }
  while (!inf.get(c).eof()) {
    if (inComment) {
      if (c == '\n') {
        inComment = false;
        if (doneToken) { break; }
      }
    } else {
      if (c == '!') {
        if (inToken) { doneToken = true; }
        inComment = true;
      } else {
        if ((tokenSep.find(c) != string::npos) || iscntrl(c)) {
          if (inToken) { doneToken = true; break; }
        } else {
          if (!inToken) { inToken = true; }
          s += c;
        }
      }
    }
  }
  if (!doneToken) { return ERRE; }
  return NORM;
}

/*----- get all characters from a file ---------------------------------
     ignoring comments ('!' to line-end), space, and control codes
----------------------------------------------------------------------*/

int getCharsFile(string flname, string &s) {
  char c;
  int inComment = false;
  ifstream inf(flname.c_str(), ios::in | ios::binary);
  if (!inf) { return ERRE; }
  s = "";
  while (!inf.eof()) {
    inf.get(c);
    if (inComment) {
      if (c == '\n') { inComment = false; }
    } else {
      if (c == '!') { inComment = true; }
      else if (isgraph(c)) { s += c; }
    }
  }
  inf.close();
  return NORM;
}

/*----- get base name of the filename ----------------------------------*/

string getBaseName(string &s) {
  int posDot, posSlsh, posBsls, posSep = -1;
  string ws, bnam;

  posSlsh = s.rfind('/');
  posBsls = s.rfind('\\');
  if (posSlsh != string::npos) {
    if (posBsls != string::npos) {
      if (posBsls > posSlsh) { posSep = posBsls; }
      else { posSep = posSlsh; }
    } else { posSep = posSlsh; }
  } else if (posBsls != string::npos) { posSep = posBsls; }
  ws = s.substr(posSep + 1);
  if ((posDot = ws.rfind('.')) != string::npos)
    { bnam = ws.substr(0, posDot); }
  else { bnam = ws; }
  return bnam;
}

/*----- get base part of the name end with strEnd ----------------------*/
string getBasePartEndWith(string nam, string strEnd) {
  int lnam, lstre;
  string ws;

  lnam = nam.length(); lstre = strEnd.length();
  if (lnam < lstre) { return getBaseName(nam); }
  if (nam.substr(lnam - lstre, lstre) != strEnd)
    { return getBaseName(nam); }
  ws = nam.substr(0, lnam - lstre) + '.';
  return getBaseName(ws);
}

/*----- G.C.M. and L.C.M. ----------------------------------------------*/

int getGCM(int x, int y) {
  int mx, im, gcm;
  if (x > y) { mx = x; } else { mx = y; }
  for (im = 2, gcm = 1; im <= mx; im++) {
    if (((x % im) == 0) && ((y % im) == 0)) { gcm = im; }
  }
  return gcm;
}

int getLCM(int x, int y) {
  int gcm, lcm;
  gcm = getGCM(x, y);
  lcm = x * y / gcm;
  return lcm;
}

/*----- file copy ------------------------------------------------------*/

int copyFile(string flnin, string flnout) {
  string s;
  ifstream ifs; ofstream ofs;
  ifs.open(flnin.c_str());
  if (!ifs) { cout << " Failed to read " << flnin << ".\n"; return ERRE; }
  ofs.open(flnout.c_str());
  if (!ofs) {
    cout << " Failed to write " << flnout << ".\n";
    ifs.close(); return ERRE;
  }
  while (!ifs.eof()) {
    if (getOneLine(ifs, s) == ERRE) { break; }
    ofs << s << endl;
  }
  ofs.close();
  ifs.close();
  return NORM;
}

/*----- add 2 files into a file ----------------------------------------*/

int addToFile(string flnin1, string flnin2, string flnout) {
  string s;
  ifstream ifs1, ifs2; ofstream ofs;
  ofs.open(flnout.c_str());
  if (!ofs) { cout << " Failed to write " << flnout << ".\n"; return ERRE; }
  ifs1.open(flnin1.c_str());
  if (!ifs1) {
    cout << " Failed to read " << flnin1 << ".\n";
    ofs.close(); return ERRE;
  }
  while (!ifs1.eof()) {
    if (getOneLine(ifs1, s) == ERRE) { break; }
    ofs << s << endl;
  }
  ifs1.close();
  ifs2.open(flnin2.c_str());
  if (!ifs2) {
    cout << " Failed to read " << flnin2 << ".\n";
    ofs.close(); return ERRE;
  }
  while (!ifs2.eof()) {
    if (getOneLine(ifs2, s) == ERRE) { break; }
    ofs << s << endl;
  }
  ifs2.close();
  ofs.close();
  return NORM;
}

/*----- add 3 files into a file ----------------------------------------*/

int add3Files(string flnin1, string flnin2, string flnin3, string flnout) {
  string s;
  ifstream ifs1, ifs2, ifs3; ofstream ofs;
  ofs.open(flnout.c_str());
  if (!ofs) { cout << " Failed to write " << flnout << ".\n"; return ERRE; }

  ifs1.open(flnin1.c_str());   // ----- file 1 -----
  if (!ifs1) {
    cout << " Failed to read " << flnin1 << ".\n";
    ofs.close(); return ERRE;
  }
  while (!ifs1.eof()) {
    if (getOneLine(ifs1, s) == ERRE) { break; }
    ofs << s << endl;
  }
  ifs1.close();

  ifs2.open(flnin2.c_str());   // ----- file 2 -----
  if (!ifs2) {
    cout << " Failed to read " << flnin2 << ".\n";
    ofs.close(); return ERRE;
  }
  while (!ifs2.eof()) {
    if (getOneLine(ifs2, s) == ERRE) { break; }
    ofs << s << endl;
  }
  ifs2.close();

  ifs3.open(flnin3.c_str());   // ----- file 3 -----
  if (!ifs3) {
    cout << " Failed to read " << flnin3 << ".\n";
    ofs.close(); return ERRE;
  }
  while (!ifs3.eof()) {
    if (getOneLine(ifs3, s) == ERRE) { break; }
    ofs << s << endl;
  }
  ifs3.close();

  ofs.close();
  return NORM;
}

// =========================================================================
//        library functions and classes for KUCRS
// =========================================================================

void showRevision() { cout << KUCRS_REV << endl; }         // showRevision

void printHeadingInfo(ostream &os) {                   // printHeadingInfo
  os << "!\n";
  os << "! generated by " << KUCRS_REV << " " << COPYR_STR << endl;
  os << "!  on " << strCurDate() << ".\n";
  os << "!\n";
}

/*===== 'libPath' class ==================================================*/

// default constructor initizlize the object with ENVVARNAM

libPath::libPath() { init(ENVVARNAM); }                // --- constructors

libPath::libPath(const string &namEnvVar) { init(namEnvVar); }

void libPath::init(const string &namEnvVar) {                  // --- init
  char *pev;
  int lenp;
  if ((pev = getenv(namEnvVar.c_str())) == NULL) { pathStr = ""; }
  else {
    pathStr = pev;
    if ((lenp = pathStr.length()) > 0) {
      if (pathStr[lenp - 1] != '/') { pathStr += '/'; }
    }
  }
}

// ------------------------------------------------------------ [fullPath]
string libPath::fullPath(string libname) {
  string rstr = pathStr;
  rstr += libname;
  return rstr;
}

/*===== 'macroMap' and related classes ===================================*/

macroMap::macroMap()                                   // --- constructors
  { clear(); alwdChars = ""; alwdNonTop = ""; }

macroMap::macroMap(string alwdch, string anontop) { init(alwdch, anontop); }

void macroMap::init(string alwdch, string anontop)             // --- init
  { clear(); alwdChars = alwdch; alwdNonTop = anontop; }

// ------------------------------------------------------------ [loadFrom]
int macroMap::loadFrom(string flname, libPath &libp) {
  string keyStr, valStr;
  if (!empty()) { clear(); }
  ifstream ifs(libp.fullPath(flname).c_str());
  if (!ifs) { return ERRE; }
  while (getTokenFrom(ifs, keyStr) == NORM) {
    if (getTokenFrom(ifs, valStr) == ERRE) { return ERRE; }
    insert(make_pair(keyStr, valStr));
  }
  ifs.close();
  cout << size() << " macros loaded from [" << flname << "]\n";
  return NORM;
}

// ---------------------------------------------------------- [getSubsVal]
string macroMap::getSubsVal(string key) {
  string rets = "";
  iterator p = find(key);
  if (p != end()) { rets = p->second; }
  return rets;
}

// --------------------------------------------------------- [substMacros]
int macroMap::substMacros(string &tgs) {
  int mfound;
  long lentgs, pstt = 0, pend, lenkeys, lenvals;
  string keys, vals;
  do {
    lentgs = tgs.length();
    for (mfound = false; pstt < lentgs; pstt++)
      { if (isAllowedTop(tgs[pstt])) { mfound = true; break; } }
    if (!mfound) { break; }
    for (pend = pstt + 1; pend < lentgs; pend++)
      { if (!isAllowed(tgs[pend])) { break; } }
    lenkeys = pend - pstt;
    if (lenkeys > 0) {
      keys.assign(tgs, pstt, lenkeys);
      vals = getSubsVal(keys);
      if (vals == "") {
        cout << "undefined macro: [" << keys << "].\n"; return ERRE;
      } else {
        tgs.replace(pstt, lenkeys, vals);
        lenvals = vals.length();
        pend = pend + lenvals - lenkeys;
      }
    }
    pstt = pend;
  } while (mfound);
  return NORM;
}

// ----------------------------------------------------------- [isAllowed]
int macroMap::isAllowed(char c) {
  if (alwdChars.find(c) != string::npos) { return true; }
  if (alwdNonTop.find(c) != string::npos) { return true; }
  return false;
}

// -------------------------------------------------------- [isAllowedTop]
int macroMap::isAllowedTop(char c) {
  if (alwdChars.find(c) != string::npos) { return true; }
  return false;
}

/*===== 'outStrms' class =================================================*/

int outStrms::open(string fnt, string fnr, string fnm, string fnl) {
  thm.open(fnt.c_str()); rxn.open(fnr.c_str());
  mol.open(fnm.c_str()); mll.open(fnl.c_str());
  if ((!thm) || (!rxn) || (!mol) || (!mll))
    { cout << "Error opening output files.\n"; return ERRE; }
  return NORM;
}

void outStrms::close()
  { thm.close(); rxn.close(); mol.close(); mll.close(); }

